@@ -1,4 +1,4 @@
-__version__ = "0.0.17"
+__version__ = "0.0.19"
 __keywords__ = ["python cgi apache"]
 
 
@@ -10,5 +10,11 @@ __keywords__ = ["python cgi apache"]
 #     os._exit(1)
 
 
-from .core import _SERVER, _GET, _POST, _SESSION, _COOKIE, _HEADERS, set_status, set_header, execute, print, main, log, log_construct
+from .core import (
+    _SERVER, _GET, _POST, _SESSION, _COOKIE, _HEADERS,
+    set_status, set_header,
+    execute, print, flush, main,
+    log, log_construct,
+    should_return_304
+)
 
